# -*- coding: utf-8 -*-

import urllib.parse
import difflib

import requests
import wsgiadapter
import lxml.html
from lxml.html import builder as E
from parsel.csstranslator import HTMLTranslator
from werkzeug.http import parse_options_header


class MissingFieldError(Exception):
    def __init__(self, field_name):
        self.field_name = field_name

    def __str__(self):
        return "MissingFieldError: Field {} cannot be found".format(
            self.field_name
        )


class MissingFormError(Exception):
    def __str__(self):
        return "MissingFormError: No form found on the page"


class NotFound(Exception):
    def __init__(self, response):
        self.response = response

    def __str__(self):
        return "{} returned a 404".format(self.response)


class MethodNotAllowed(Exception):
    def __init__(self, response):
        self.response = response

    def __str__(self):
        return "{} returned a 405".format(self.response)


def print_quick_pprint_diff(item1, item2):
    if isinstance(item1, bytes):
        item1 = item1.decode("utf-8")
    if isinstance(item2, bytes):
        item2 = item2.decode("utf-8")
    diff = difflib.unified_diff(item1.split('\n'), item2.split('\n'))
    for line in list(diff):
        print(line)


def parse_html(html):
    result = lxml.html.fromstring(html)
    output = lxml.html.tostring(result).decode("utf-8")
    try:
        assert output.strip() == html.strip()
    except AssertionError:
        print_quick_pprint_diff(html, output)
        raise
    return result


class XPath(str):
    pass


class BaseAgent(object):
    def __init__(self, base_url):
        self.base_url = base_url
        self.session = requests.Session()
        self.session.hooks = {'response': self.make_response}

    def get(self, url, data=None, **kwargs):
        url = urllib.parse.urljoin(self.base_url, url)
        response = self.session.get(
            url, params=data, allow_redirects=False, **kwargs
        )
        if response.status_code == 404:
            raise NotFound(response)
        return response

    def post(self, url, data=None, **kwargs):
        url = urllib.parse.urljoin(self.base_url, url)
        response = self.session.post(
            url, data=data, allow_redirects=False, **kwargs
        )
        if response.status_code == 405:
            raise MethodNotAllowed(response)
        return response

    def make_response(self, response, **kwargs):
        return Response(response=response, agent=self)


class WSGIAgent(BaseAgent):
    def __init__(self, wsgi_app, base_url="http://example.com/"):
        super(WSGIAgent, self).__init__(base_url)
        self.app = wsgi_app
        self.session.mount(self.base_url, wsgiadapter.WSGIAdapter(self.app))


class Response(object):
    def __init__(self, response, agent):
        self.response = response
        self.agent = agent
        if self.mime_type == "text/html" and self.content:
            self.lxml = parse_html(self.content)
        else:
            self.lxml = None

    def __getattr__(self, key):
        return getattr(self.response, key)

    def __repr__(self):
        return "<Request {} {}>".format(self.status_code, self.path)

    @property
    def path(self):
        return urllib.parse.urlparse(self.url).path

    @property
    def mime_type(self):
        content_type = self.headers.get('Content-Type')
        if content_type:
            return parse_options_header(content_type)[0]

    @property
    def charset(self):
        return parse_options_header(self.headers['Content-Type'])[1].get(
            'charset', ""
        )

    def to_string(self, charset=None):
        if not charset:
            charset = self.charset
        return lxml.html.tostring(self.lxml).decode(charset)

    def one(self, selector):
        if not isinstance(selector, XPath):
            selector = HTMLTranslator().css_to_xpath(selector)
        els = self.lxml.xpath(selector)
        assert len(els) == 1, "Length is {}".format(len(els))
        return els[0]

    def has_one(self, selector):
        selector = HTMLTranslator().css_to_xpath(selector)
        els = self.lxml.xpath(selector)
        return len(els) == 1

    def has_text(self, text):
        selector = "*:contains({})".format(repr(text))
        selector = HTMLTranslator().css_to_xpath(selector)
        els = self.lxml.xpath(selector)
        return len(els) > 0

    def all(self, selector):
        selector = HTMLTranslator().css_to_xpath(selector)
        els = self.lxml.xpath(selector)
        return tuple(els)

    def click(self, selector=None, contains=None, index=None):
        if contains:
            selector = "a:contains('{}')".format(contains)
        els = self.all(selector)
        if index is None:
            assert len(els) > 0, "No matching links"
            assert len(els) < 2, "Too many matching links"
            index = 0
        el = els[index]
        url = el.attrib['href']
        return self.agent.get(url)

    def get_form(self, selector="form", index=None):
        els = self.all(selector)
        if index is None:
            assert len(els) > 0, "No matching forms"
            assert len(els) < 2, "Too many matching forms"
            index = 0
        el = els[index]
        return Form(self, el)

    def follow(self):
        assert 300 <= self.status_code < 400, self.status_code
        location = self.headers['Location']
        response = self.agent.get(location)
        return response


class Form(object):
    def __init__(self, response, element):
        self.response = response
        self.element = element

    @property
    def method(self):
        return self.element.attrib.get('method', 'GET').upper()

    @property
    def action(self):
        return self.element.attrib.get('action')

    def __getitem__(self, key):
        return self.element.fields[key]

    def __setitem__(self, key, value):
        if isinstance(value, int):
            value = str(value)
        self.element.fields[key] = value

    def check(self, label):
        selector = "label:contains({})".format(repr(label))
        label_els = self.element.cssselect(selector)
        assert len(label_els) == 1
        label_el = label_els[0]
        if label_el.for_element is not None:
            el = label_el.for_element
        else:
            el = label_el.cssselect("input")[0]
        el.checked = not el.checked

    def select(self, field_name, value, force=False):
        field = self.element.cssselect('''select[name={}]'''.format(field_name))[0]
        if not value in field.value_options:
            field.append(
                E.OPTION(value=str(value))
            )
        self[field_name] = value

    def set(self, field_name, value):
        if field_name in self.element.fields:
            self[field_name] = value
        else:
            self.element.append(
                E.INPUT(type="hidden", name=field_name, value=str(value))
            )

    def submit_data(self):
        data = dict(self.element.form_values())
        for field in self.element.inputs:
            if (field.attrib.get('type') == 'submit') and field.attrib.get('name'):
                data[field.attrib['name']] = field.value
        return data

    def submit(self, data=None, extra=None):
        if data is None:
            data = self.submit_data()
        if extra:
            data.update(extra)
        action = urllib.parse.urljoin(self.response.url, self.action)
        func = getattr(self.response.agent, self.method.lower())
        response = func(action, data=data)
        return response

    def to_string(self, charset=None):
        if not charset:
            charset = self.response.charset
        if self.element:
            return lxml.html.tostring(self.element).decode('ascii')
        else:
            return None



# -*- coding: utf-8 -*-

import unittest.mock

import nose
import pyjade
import requests_mock

from testino import Response, XPath, WSGIAgent, BaseAgent, MissingFieldError, MissingFormError

document = pyjade.simple_convert(
    '''
html
  body
    div#foo
      p This is foo
    div#bar
      p This is bar
    a#bumble(href="/bumble")
      button Bumble
    a#famble(href="/famble")
      button Famble
''')

form_document = pyjade.simple_convert(
    '''
html
  body
    form#one(action="/result_page_1")
      input(name="flibble_1")
    form#two(action="/result_page_2")
      input(name="flibble_2")
''')


def wsgi_app(env, start_response):
    start_response('200 OK', [('Content-Type', 'text/html')])
    return [b"This is a WSGI app"]


def test_WSGIAgent():
    agent = WSGIAgent(wsgi_app)
    response = agent.get("/")
    assert response.content == b"This is a WSGI app"


@requests_mock.mock()
def test_BaseAgent(mock_requests):
    mock_requests.get("http://example.com/foo", text='This is not a WSGI app')
    agent = BaseAgent("http://example.com")
    response = agent.get("/foo")
    assert response.content == b"This is not a WSGI app"


@requests_mock.mock()
def test_headers(mock_requests):
    mock_requests.get("http://example.com/foo", text='This is not a WSGI app')
    agent = BaseAgent("http://example.com")
    agent.get("/foo", headers={"Foo": "Bar"})
    assert mock_requests.request_history[0].headers['Foo'] == 'Bar'


class StubResponse(object):
    def __init__(self, content):
        self.content = content
        self.url = "http://example.com/flibble"
        self.headers = {"Content-Type": "text/html; charset=utf-8"}
        self.status_code = 999


class TestResponse(unittest.TestCase):
    def setUp(self):
        self.mock_agent = unittest.mock.Mock()
        self.response = Response(
            StubResponse(document), agent=self.mock_agent)

    def test_repr(self):
        assert str(self.response) == "<Request 999 /flibble>"

    def test_path(self):
        path = self.response.path
        assert path == "/flibble"

    def test_mime_type(self):
        assert self.response.mime_type == "text/html"

    def test_charset(self):
        assert self.response.charset == "utf-8"

    def test_one(self):
        el = self.response.one("div#foo")
        assert el.text_content().strip() == "This is foo"

    def test_one_fails(self):
        with nose.tools.assert_raises(AssertionError):
            self.response.one("div#fumble")

    def test_one_xpath(self):
        el = self.response.one(XPath("//div[@id='foo']"))
        assert el.text_content().strip() == "This is foo"

    def test_one_fails_xpath(self):
        with nose.tools.assert_raises(AssertionError):
            self.response.one(XPath("//div[@id='fumble']"))

    def test_has_one(self):
        assert self.response.has_one("div#foo")

    def test_has_one_fails(self):
        assert not self.response.has_one("div#fumble")

    def test_all(self):
        els = self.response.all("div")
        assert len(els) == 2
        assert isinstance(els, tuple)

    def test_has_text(self):
        assert self.response.has_text("This is foo")

    def test_has_text_fails(self):
        assert not self.response.has_text("Say hello to Mr Flibble")

    def test_click_contains(self):
        self.response.click(contains="Bumble")
        expected_calls = [unittest.mock.call.get('/bumble')]
        assert self.mock_agent.mock_calls == expected_calls

    def test_click_contains_index(self):
        self.response.click(contains="ble", index=1)
        expected_calls = [unittest.mock.call.get('/famble')]
        assert self.mock_agent.mock_calls == expected_calls

    def test_click_id(self):
        self.response.click("#bumble")
        expected_calls = [unittest.mock.call.get('/bumble')]
        assert self.mock_agent.mock_calls == expected_calls

    def test_click_no_links(self):
        with nose.tools.assert_raises(AssertionError) as e:
            self.response.click(contains="FlubNuts")
        expected_error = "No matching links"
        assert str(e.exception) == expected_error, e.exception

    def test_click_too_many_links(self):
        with nose.tools.assert_raises(AssertionError) as e:
            self.response.click(contains="ble")
        expected_error = "Too many matching links"
        assert str(e.exception) == expected_error, e.exception

    def test_get_form_no_form(self):
        with nose.tools.assert_raises(AssertionError) as e:
            self.response.get_form()
        expected_error = "No matching forms"
        assert str(e.exception) == expected_error, e.exception

    def test_get_from_too_many_forms(self):
        response = Response(StubResponse(form_document), agent=self.mock_agent)
        with nose.tools.assert_raises(AssertionError) as e:
            response.get_form()
        expected_error = "Too many matching forms"
        assert str(e.exception) == expected_error, e.exception

    def test_broken_html(self):
        broken_html = """<div>"""
        response = Response(StubResponse(broken_html), agent=self.mock_agent)
        print(response.to_string())
        assert broken_html == response.to_string()
        



from setuptools import setup

data = dict(
    name='testino',
    version='0.3.11',
    description="Test WSGI applications using lxml",
    long_description="",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development :: Testing',
    ],
    keywords='',
    author='Ed Singleton',
    author_email='singletoned@gmail.com',
    url='https://github.com/Singletoned/testino',
    license='BSD',
    py_modules=['testino'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "lxml>=3.7",
        "werkzeug>=0.12",
        "cssselect>=1",
        "parsel>=1.1",
        "requests>=2",
        "requests-wsgi-adapter>=0.2",
    ],
    entry_points="""
    # -*- Entry points: -*-
    """,
)

if __name__ == '__main__':
    setup(**data)
